/* ====================================================================================================================

  The copyright in this software is being made available under the License included below.
  This software may be subject to other third party and contributor rights, including patent rights, and no such
  rights are granted under this license.

  Copyright (c) 2011, TECHNICOLOR
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of SAMSUNG ELECTRONICS CO., LTD. nor the name of the BRITISH BROADCASTING CORPORATION
      may be used to endorse or promote products derived from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

/** \file     TComWeightPrediction.h
    \brief    weighting prediction class (header)
*/

#ifndef __TCOMWEIGHTPREDICTION__
#define __TCOMWEIGHTPREDICTION__


// Include files
#include "TComPic.h"
#include "TComMotionInfo.h"
#include "TComPattern.h"
#include "TComTrQuant.h"

#ifdef WEIGHT_PRED

// ====================================================================================================================
// Class definition
// ====================================================================================================================
/// weighting prediction class
class TComWeightPrediction
{
  wpScalingParam  m_wp0[3], m_wp1[3];
  Int             m_ibdi;

public:
  TComWeightPrediction();

  Void  getWpScaling( TComDataCU* pcCU, Int iRefIdx0, Int iRefIdx1, wpScalingParam *&wp0 , wpScalingParam *&wp1 , Int ibdi=(g_uiBitDepth+g_uiBitIncrement));

  Void  addWeightBi( TComYuv* pcYuvSrc0, TComYuv* pcYuvSrc1, UInt iPartUnitIdx, UInt iWidth, UInt iHeight, wpScalingParam *wp0, wpScalingParam *wp1, TComYuv* rpcYuvDst, Bool bRound );
  Void  addWeightUni( TComYuv* pcYuvSrc0, UInt iPartUnitIdx, UInt iWidth, UInt iHeight, wpScalingParam *wp0, TComYuv* rpcYuvDst );

  Void  xWeightedPredictionUni( TComDataCU* pcCU, TComYuv* pcYuvSrc, UInt uiPartAddr, Int iWidth, Int iHeight, RefPicList eRefPicList, TComYuv*& rpcYuvPred, Int iPartIdx );
  Void  xWeightedPredictionBi( TComDataCU* pcCU, TComYuv* pcYuvSrc0, TComYuv* pcYuvSrc1, Int iRefIdx0, Int iRefIdx1, UInt uiPartIdx, Int iWidth, Int iHeight, TComYuv* rpcYuvDst );

  __inline  Pel xClip  ( Int x );
  __inline  Pel weightBidir( Int w0, Pel P0, Int w1, Pel P1, Int round, Int shift, Int offset);
  __inline  Pel weightUnidir( Int w0, Pel P0, Int round, Int shift, Int offset);
  __inline  Pel weightBidirOff( Int w0, Pel P0, Int w1, Pel P1, Int round, Int shift, Int offset, Int uvOff);
  __inline  Pel weightUnidirOff( Int w0, Pel P0, Int round, Int shift, Int offset, Int uvOff);

};

inline  Pel TComWeightPrediction::xClip( Int x )
{ 
#if HIGH_ACCURACY_BI
  Int max = (1 << 14);
#else
  Int max = (Int)g_uiIBDI_MAX;
#endif

  Pel pel = (Pel)( (x < 0) ? 0 : (x > max) ? max : x );

  return( pel );
}


inline  Pel TComWeightPrediction::weightBidir( Int w0, Pel P0, Int w1, Pel P1, Int round, Int shift, Int offset)
{
  return xClip( ( (w0*P0 + w1*P1 + round) >> shift ) + offset );
}

inline  Pel TComWeightPrediction::weightUnidir( Int w0, Pel P0, Int round, Int shift, Int offset) 
{
  return xClip( ( (w0*P0 + round) >> shift ) + offset );
}

inline  Pel TComWeightPrediction::weightBidirOff( Int w0, Pel P0, Int w1, Pel P1, Int round, Int shift, Int offset, Int uvOff)
{
  return xClip( ( (w0*((Int)P0-uvOff) + w1*((Int)P1-uvOff) + round) >> shift ) + uvOff + offset );
}

inline  Pel TComWeightPrediction::weightUnidirOff( Int w0, Pel P0, Int round, Int shift, Int offset, Int uvOff)
{
  return xClip( ( (w0*((Int)P0-uvOff) + round) >> shift ) + uvOff + offset );
}

#endif  // WEIGHT_PRED

#endif


