//
// C++ Interface: WeightPredTest
//
// Description: Class for loading Weightd Prediction parameters.
//
//
// Authors:  <Bordes Philippe>, (C) 2011
//           Technicholor Inc.
//
//               10/3/2011
//
// Copyright: All rights reserved Technicholor Corp. Res.
//

#include "../TLibCommon/TypeDef.h"
#include "../TLibCommon/TComSlice.h"
#include "WeightPredTest.h"

#ifdef WEIGHT_PRED

#define LINE_SIZE_MAX  400

Char            *WeightPredTest::m_pchWpScalingFile=NULL;
Int							WeightPredTest::m_pocStart=(-1);
Int							WeightPredTest::m_pocEnd=(-1);              // Starting and Ending Fading times
Bool					  WeightPredTest::m_bDataLoaded=false;

Bool  read2Words(FILE *fic, Char *name, Char word[2*MAX_NUM_REF][LINE_SIZE_MAX], Int nbWord=1)
{
  Char  line[LINE_SIZE_MAX], *p;
  Int   n=0;

  name[0]     = '\0';

  if ( (p=fgets(line, LINE_SIZE_MAX, fic)) ) {
    if ( (p = strtok(p," \t\n")) ) {
      strcpy(name,p);
      while ( (p = strtok(NULL," ")) ) 
        strcpy(word[n++],p);
    }
    else {
      word[n][0] = '\0';
      return(true);
    }
  }
  else
    return(false);

  word[n][0] = '\0';

  return(true);
}

Void  WeightPredTest::resetPredTable(wpScalingParam m_weightPredTable[100][2][MAX_NUM_REF][3])
{
  m_pocStart  = 0;
  m_pocEnd    = 0;

  for ( int iPoc=0 ; iPoc<WP_DATA_SIZE ; iPoc++ )
  for ( int e=0 ; e<2 ; e++ )
  for ( int i=0 ; i<MAX_NUM_REF ; i++ )
  for ( int yuv=0 ; yuv<3 ; yuv++ ) {
    m_weightPredTable[iPoc][e][i][yuv].bPresentFlag       = false;
    m_weightPredTable[iPoc][e][i][yuv].uiLog2WeightDenom  = 0;
    m_weightPredTable[iPoc][e][i][yuv].uiLog2WeightDenom  = 0;
    m_weightPredTable[iPoc][e][i][yuv].iWeight            = 1;
    m_weightPredTable[iPoc][e][i][yuv].iOffset            = 0;
  }
}

Bool  WeightPredTest::readWeightPredFile(wpScalingParam m_weightPredTable[100][2][MAX_NUM_REF][3])
{
  if ( !m_pchWpScalingFile )  return false;
  Char  line[LINE_SIZE_MAX], name[LINE_SIZE_MAX], word[2*MAX_NUM_REF][LINE_SIZE_MAX];
  FILE  *fic = fopen(m_pchWpScalingFile, "r");
  Char  *p;

  p=fgets(line, LINE_SIZE_MAX, fic);
  sscanf(line, "%d", &m_pocStart);
  p=fgets(line, LINE_SIZE_MAX, fic);
  sscanf(line, "%d", &m_pocEnd);

  if ( (m_pocEnd - m_pocStart + 1) > WP_DATA_SIZE ) {
    printf("WeightPredTest::readWeightPredFile():\t'%s'contains more than %d samples (exceeds WP_DATA_SIZE value).\n", m_pchWpScalingFile, WP_DATA_SIZE);
    exit(0);
  }

  Int   iRef[2]={0,0}, iPoc=0;
  Int   eRef=0, yuv=0;
  Bool  bFound=false;

  Bool  bCont = true;
  while ( bCont ) {

    if ( !read2Words(fic, name, word) )  { bCont = false; continue; }
    if ( strcmp(name,"poc")==0 && strlen(word[0]) ) {
      iPoc    = atoi(word[0]) - m_pocStart;
      iRef[0] = 0;
      iRef[1] = 0;
      eRef    = 0;

      for ( int e=0 ; e<2 ; e++ )
      for ( int i=0 ; i<MAX_NUM_REF ; i++ )
      for ( yuv=0 ; yuv<3 ; yuv++ )
        m_weightPredTable[iPoc][e][i][yuv].bPresentFlag = false;

      continue;
    }
    
    if ( strcmp(name,"iref0")==0 && strlen(word[0]) ) {  iRef[0] = atoi(word[0]); }
    if ( strcmp(name,"iref1")==0 && strlen(word[0]) ) {  iRef[1] = atoi(word[0]); }

    bFound = false;

    if ( strcmp(name,"yw0")==0 && strlen(word[0]) ) {  eRef  = 0;  yuv   = 0;  bFound = true;  }
    if ( strcmp(name,"uw0")==0 && strlen(word[0]) ) {  eRef  = 0;  yuv   = 1;  bFound = true;  }
    if ( strcmp(name,"vw0")==0 && strlen(word[0]) ) {  eRef  = 0;  yuv   = 2;  bFound = true;  }
    if ( strcmp(name,"yw1")==0 && strlen(word[0]) ) {  eRef  = 1;  yuv   = 0;  bFound = true;  }
    if ( strcmp(name,"uw1")==0 && strlen(word[0]) ) {  eRef  = 1;  yuv   = 1;  bFound = true;  }
    if ( strcmp(name,"vw1")==0 && strlen(word[0]) ) {  eRef  = 1;  yuv   = 2;  bFound = true;  }
    
    if ( bFound ) {
      if ( iRef[eRef] >= MAX_NUM_REF )  { bCont = false; continue; }

      m_weightPredTable[iPoc][eRef][iRef[eRef]][yuv].iWeight      = atoi(word[0]);
      m_weightPredTable[iPoc][eRef][iRef[eRef]][yuv].bPresentFlag = true;
      continue;
    }

    if ( strcmp(name,"yo0")==0 && strlen(word[0]) ) {  eRef  = 0;  yuv   = 0;  bFound = true;  }
    if ( strcmp(name,"uo0")==0 && strlen(word[0]) ) {  eRef  = 0;  yuv   = 1;  bFound = true;  }
    if ( strcmp(name,"vo0")==0 && strlen(word[0]) ) {  eRef  = 0;  yuv   = 2;  bFound = true;  }
    if ( strcmp(name,"yo1")==0 && strlen(word[0]) ) {  eRef  = 1;  yuv   = 0;  bFound = true;  }
    if ( strcmp(name,"uo1")==0 && strlen(word[0]) ) {  eRef  = 1;  yuv   = 1;  bFound = true;  }
    if ( strcmp(name,"vo1")==0 && strlen(word[0]) ) {  eRef  = 1;  yuv   = 2;  bFound = true;  }

    if ( bFound ) {
      if ( iRef[eRef] >= MAX_NUM_REF )  { bCont = false; continue; }

      m_weightPredTable[iPoc][eRef][iRef[eRef]][yuv].iOffset = atoi(word[0]);
      m_weightPredTable[iPoc][eRef][iRef[eRef]][yuv].bPresentFlag = true;
      continue;
    }

    if ( strcmp(name,"ld")==0 && strlen(word[0]) ) {
      for ( int e=0 ; e<2 ; e++ )
        for ( int i=0 ; i<MAX_NUM_REF ; i++ ) {
          m_weightPredTable[iPoc][e][i][0].uiLog2WeightDenom = atoi(word[0]);
          if ( !m_weightPredTable[iPoc][e][i][0].bPresentFlag ) {
            m_weightPredTable[iPoc][e][i][0].iWeight = (1 << m_weightPredTable[iPoc][e][i][0].uiLog2WeightDenom);
            m_weightPredTable[iPoc][e][i][0].iOffset = 0;
          }
        }
    }

    if ( strcmp(name,"cd")==0 && strlen(word[0]) ) {
      for ( int e=0 ; e<2 ; e++ )
      for ( int i=0 ; i<MAX_NUM_REF ; i++ )
        for ( yuv=1 ; yuv<3 ; yuv++ ) {
          m_weightPredTable[iPoc][e][i][yuv].uiLog2WeightDenom = atoi(word[0]);
          if ( !m_weightPredTable[iPoc][e][i][yuv].bPresentFlag ) {
            m_weightPredTable[iPoc][e][i][yuv].iWeight = (1 << m_weightPredTable[iPoc][e][i][yuv].uiLog2WeightDenom);
            m_weightPredTable[iPoc][e][i][yuv].iOffset = 0;
          }
        }
    }

  }

  fclose(fic);
  return(true);
}
	
Bool  WeightPredTest::loadWeightPredTable(TComSlice *slice, int poc)
{
  static  wpScalingParam	m_weightPredTable[WP_DATA_SIZE][2][MAX_NUM_REF][3];
  
  Bool  implicit=false;
  if ( slice->getSliceType()==B_SLICE && slice->getPPS()->getWPBiPredIdc()==2 )
    implicit = true;

  if ( !m_bDataLoaded ) {
    resetPredTable(m_weightPredTable);
    if ( !implicit )
      readWeightPredFile(m_weightPredTable);
    m_bDataLoaded = true;
  }

  if ( poc < m_pocStart ) return(false);
  if ( poc > m_pocEnd )   return(false);

  Int iPoc = poc - m_pocStart;

  printf("\nWeightPredTest::getWpData(poc=%d):\n", poc);

  slice->setWpScaling( m_weightPredTable[implicit ? (0) : (iPoc)] );

  return(true);
}

Void  WeightPredTest::writeRefPicList(TComSlice *pcSlice)
{
  static  FILE  *sFic=NULL;

  if ( !sFic ) {
    sFic = fopen("refPicList.txt", "w");
  }

  // Display Ref. Pict. List
  printf("\nPOC[%d] -----------------------------\n", pcSlice->getPOC() );
  fprintf(sFic, "poc %d -----------------\n",           pcSlice->getPOC() );
  for ( Int e=0 ; e<2 ; e++ ) {
    RefPicList eRefPicList;
    switch( e ) {
      case  0 : eRefPicList = REF_PIC_LIST_0; break;
      case  1 : eRefPicList = REF_PIC_LIST_1; break;
      case  2 : eRefPicList = REF_PIC_LIST_C; break;
      default : eRefPicList = REF_PIC_LIST_0; break;  // just for compilation warnings...
    };
    for ( Int iRef=0 ; iRef<pcSlice->getNumRefIdx(eRefPicList) ; iRef++ ) {
      printf("poc_%d:\t%d\n", e, pcSlice->getRefPOC(eRefPicList, iRef) );
      fprintf(sFic, "poc_%d:\t%d\n", e, pcSlice->getRefPOC(eRefPicList, iRef) );
    }
  }

  fflush(sFic);
}

#endif	// WEIGHT_PRED

