/* ====================================================================================================================

  The copyright in this software is being made available under the License included below.
  This software may be subject to other third party and   contributor rights, including patent rights, and no such
  rights are granted under this license.

  Copyright (c) 2011, TECHNICOLOR
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * The name of TECHNICOLOR
      may not be used to endorse or promote products derived from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "Param.h"

typedef       unsigned char       byte;

#define LINE_SIZE_MAX       (200)
#define MIN(a,b)            ((a)<(b)?(a):(b))
#define MAX(a,b)            ((a)>(b)?(a):(b))
#define CLAMP(val,min,max)  (MIN(MAX(val,min),max))

#define BitIncrement  (4)

typedef struct {
  double  K0, K1;
} ScaleParam;

void  error()
{
  printf("error.\n");
  exit(0);
}

void  black(byte *ima, int sizeX, int sizeY)
{
  int nbp=sizeX*sizeY;

  for ( int n=0 ; n<nbp ; n++ ) ima[n] = 0;

  ima += nbp;
  nbp /= 4;

  for ( int n=0 ; n<nbp ; n++ ) ima[n] = 128;
  
  ima += nbp;

  for ( int n=0 ; n<nbp ; n++ ) ima[n] = 128;
}

void  setWpScale(ScaleParam *sc, double K0, double K1)
{
  sc->K0 = K0;
  sc->K1 = K1;
}

int computeBestLogWD_2(double K0, double K1)
{
  int     iK;
  double  bestSumErr=99999999.9;
  int     bestLogWD=0;

  for ( int logwd=0 ; logwd<8 ; logwd++ ) {
    double  sumErr=0.0;
    int     cpt=0;

    for ( int luma=10 ; luma<=250 ; luma++ ) {

      int iluma = luma << BitIncrement;

      if ( K0 >= 0.0 ) {
        iK  = (int)( K0 * (1 << logwd) );

        int ival = (iK*iluma)>>logwd;
        sumErr += fabs( K0*iluma - ival ) / (double)iluma; cpt++;
      }
      if ( K1 >= 0.0 ) {
        iK  = (int)( K1 * (1 << logwd) );

        int ival = (iK*iluma)>>logwd;
        sumErr += fabs( K1*iluma - ival ) / (double)iluma; cpt++;
      }

    }

    if ( sumErr < bestSumErr ) {
      bestSumErr  = sumErr;
      bestLogWD   = logwd;
    }

  }

  return(bestLogWD);
}

void  fwriteWpScale_logDW(FILE *ficWpScale, int LOGWD, int poc)
{
  fprintf(ficWpScale, "poc %d -----------------\n", poc);
  fprintf(ficWpScale, "ld %d \n", LOGWD);
  fprintf(ficWpScale, "cd %d \n", LOGWD);
}

void  fwriteWpScale1(FILE *ficWpScale, int LOGWD, double K, int eRef, int iRef, int useOffsetUV, int noWeightUV)
{
  int one=1;
  int iK        = (int)( K * (double)(one << LOGWD) );
  K = (double)iK / (double)(one << LOGWD);

  double  Kuv=K;
  if ( noWeightUV ) Kuv = 1.0;
  int offsetUV  = ((128*(1.0-Kuv))>0.0) ? (int)(128*(1.0-Kuv) + 0.5) :  (int)(128*(1.0-Kuv) - 0.5) ;
  if ( !useOffsetUV )
    offsetUV = 0;

  fprintf(ficWpScale, "iref%d %d \n", eRef, iRef);

  fprintf(ficWpScale, "yw%d %d \n", eRef, iK);
  if ( noWeightUV ) {
    fprintf(ficWpScale, "uw%d %d \n", eRef, (one<<LOGWD));
    fprintf(ficWpScale, "vw%d %d \n", eRef, (one<<LOGWD));
  }
  else {
    fprintf(ficWpScale, "uw%d %d \n", eRef, iK);
    fprintf(ficWpScale, "vw%d %d \n", eRef, iK);
  }
  fprintf(ficWpScale, "yo%d %d \n", eRef, 0);
  fprintf(ficWpScale, "uo%d %d \n", eRef, offsetUV);
  fprintf(ficWpScale, "vo%d %d \n", eRef, offsetUV);
}

bool  read2Words(FILE *fic, char *word1, char *word2)
{
  char  line[LINE_SIZE_MAX], *p;

  word1[0] = '\0';
  word2[0] = '\0';

  if ( (p=fgets(line, LINE_SIZE_MAX, fic)) ) {
    if ( (p = strtok(p," \t\n")) ) {
      strcpy(word1,p);
      if ( (p = strtok(NULL," ")) ) 
        strcpy(word2,p);
    }
    else {
      word1[0] = word2[0] = '\0';
      return(true);
    }
  }
  else
    return(false);

  return(true);
}

#define MAX_NUM_POC (150)
#define MAX_NUM_REF (4)
int pocRefPicList[MAX_NUM_POC][2][MAX_NUM_REF];
void  loadRefPicList(char *dataFile)
{
  for ( int poc=0 ; poc<MAX_NUM_POC ; poc++ )
  for ( int eRef=0 ; eRef<2 ; eRef++ )
  for ( int i=0 ; i<MAX_NUM_REF ; i++ )
    pocRefPicList[poc][eRef][i] = (-1);

  FILE  *fic = fopen(dataFile, "r");

  int poc = 0,
      iRef[2]={0,0};
  char  word1[LINE_SIZE_MAX], word2[LINE_SIZE_MAX];
  while( read2Words(fic,word1,word2) ) {
    if ( strcmp(word1,"poc") == 0 ) {
      poc = atoi(word2);
      iRef[0] = iRef[1] = 0;
    }
    if ( poc<0 || poc>=MAX_NUM_POC ) 
      continue;

    if ( strcmp(word1,"poc_0:") == 0 )  pocRefPicList[poc][0][iRef[0]++] = atoi(word2);
    if ( strcmp(word1,"poc_1:") == 0 )  pocRefPicList[poc][1][iRef[1]++] = atoi(word2);
  }

  fclose(fic);
}

int computeBestLogWD(ScaleParam *sc, int n)
{
  int     iK, ref0;
  double  bestSumErr=99999999.9;
  int     bestLogWD=0;

  for ( int logwd=0 ; logwd<8 ; logwd++ ) {
    double  sumErr=0.0;
    int     cpt=0;

    for ( int luma=10 ; luma<=250 ; luma++ ) {

      int iluma = luma << BitIncrement;

      for ( int eRef=0 ; eRef<2 ; eRef++ ) {
        int iref=0;
        while ( pocRefPicList[n][eRef][iref]>=0 && iref<MAX_NUM_REF ) {
          ref0 = pocRefPicList[n][eRef][iref];
          double  K0 = sc[n].K0 / sc[ref0].K0;
          iref++;

          iK  = (int)( K0 * (1 << logwd) );

          int ival = (iK*iluma)>>logwd;
          sumErr += fabs( K0*iluma - ival ) / (double)iluma; cpt++;
        }
      }

    } // for luma

    if ( sumErr < bestSumErr ) {
      bestSumErr  = sumErr;
      bestLogWD   = logwd;
    }

  } // for logwd

  return(bestLogWD);
}

void    Synopsis()
{
  printf("=========================================================================\n");
  printf("\tFading Tool for HM software version implementing Weighted Prediction.  \n");
  printf("=========================================================================\n");
  printf("\n");
  printf("Synopsis:\tFading -i0 seq0.yuv -o output.yuv -sx sizex -sy sizey -nb nbima -sc scenario -refPicList refPocFile -wp outWpFile [ -i1 seq1.yuv, -kmin Kmin, -offsetUV ]\n");
  printf("\n");
  printf("-i0 seq0.yuv\t'seq0.yuv' is the input 0 sequence to fade or cross-fade.\n");
  printf("-i1 seq1.yuv\t'seq1.yuv' is the input 1 sequence to cross-fade.\n");
  printf("-o output.yuv\t'output.yuv' is the output fade (or cross-fade) sequence.\n");
  printf("-wp outWpFile\t'outWpFile' contains weighted prediction explicit input parameters for HM.\n");
  printf("-sx sizex -sy sizey\tinput and output sequences images sizes.\n");
  printf("-sc scenario\teither FADE_OUT_FADE_IN or CROSS_FADE.\n");
  printf("-refPicList refPocFile\tFile containing the POC of reference frames.\n");
  printf("-kmin Kmin\tMinimum Fade weight value in FADE_OUT_FADE_IN scenario.\n");
  printf("-offsetUV\tAdd Chroma offsets.\n");
  printf("-noWeightUV\tUse default weight (1.) for chroma.\n");
  printf("\n");
}

int	main(int argc, char **argv)
{
  Synopsis();
  double  Kmin=0.25;
  Param *param      = new Param();
  char  scenario[100]={"FADE_OUT_FADE_IN"};
  FILE  *ficIn[2]   = {NULL,NULL};
  FILE  *ficWpScale = NULL;
  int   wpMode=0;
  if ( param->isOption(argc, argv, "i0") ) {
    ficIn[0]  = fopen( param->readOption(argc, argv, "i0"), "rb");
    wpMode    = param->isOption(argc, argv, "i1") ? (2) : (0);
  }
  if ( param->isOption(argc, argv, "i1") ) { 
    ficIn[1]  = fopen( param->readOption(argc, argv, "i1"), "rb");
    wpMode    = param->isOption(argc, argv, "i0") ? (2) : (1);
  }

  if ( !ficIn[0] && !ficIn[1] ) {
    if ( param->isOption(argc, argv, "i0") )  printf("Opening file '%s' failed.\n", param->readOption(argc, argv, "i0"));
    if ( param->isOption(argc, argv, "i1") )  printf("Opening file '%s' failed.\n", param->readOption(argc, argv, "i1"));
    
    printf("At least one input file -i0 or -i1 should exists.\n\n");
    exit(0);
  }

  FILE  *ficOut   = fopen( param->readOption(argc, argv, "o"), "wb");
  int   sizeX     = atoi(  param->readOption(argc, argv, "sx") ),
        sizeY     = atoi(  param->readOption(argc, argv, "sy") ),
        nbIma     = atoi(  param->readOption(argc, argv, "nb") );

  if ( param->isOption(argc, argv, "kmin") )  Kmin = atof(  param->readOption(argc, argv, "kmin") );
  else                                        Kmin = 0.25;

  if ( param->isOption(argc, argv, "sc") ) strcpy(scenario, param->readOption(argc, argv, "sc") );

  int useOffsetUV = 0;
  if ( param->isOption(argc, argv, "offsetUV") )  useOffsetUV = 1;
  else                                            useOffsetUV = 0;

  int noWeightUV = 0;
  if ( param->isOption(argc, argv, "noWeightUV") )  noWeightUV = 1;
  else                                              noWeightUV = 0;

#define REF_PIC_LIST()  (param->isOption(argc, argv, "refPicList"))
  if ( REF_PIC_LIST() )
    loadRefPicList( param->readOption(argc, argv, "refPicList") );

  ScaleParam  *sc = (ScaleParam *)malloc(sizeof(ScaleParam)*nbIma);

  int nbpLuma   = sizeX*sizeY;
  int nbpChroma = (sizeX>>1)*(sizeY>>1);

  byte  *ima[3]   = {NULL,NULL,NULL};
  for ( int i=0 ; i<3 ; i++ )
    ima[i]        = (byte *)malloc(sizeof(byte)*(nbpLuma+2*nbpChroma));

  if ( param->isOption(argc, argv, "wp") ) {
    ficWpScale = fopen( param->readOption(argc, argv, "wp"), "w");
    fprintf(ficWpScale, "1\t// pocStart\n%d\t// pocEnd\n\n", nbIma-2);
  }

  // Computation of the absolute Weighting values :
  //-----------------------------------------------
  for ( int n=0 ; n<nbIma ; n++ ) {

    double  dvalue;
    double  K0=1.0, K1=0.0;
    if ( strcmp(scenario,"FADE_OUT")==0 || strcmp(scenario,"CROSS_FADE")==0 ) {
      //  Linear Fade Sequence-to-Black
      K0 = 1.0 - (1.0-Kmin)*((double)n/(double)(nbIma-1));
      K1 = 1.0 - K0;
    }
    else if ( strcmp(scenario,"FADE_OUT_FADE_IN")==0 ) {
      //  1st half, Linear Fade Sequence-to-Black, 2nd half Linear Fade Black-to-Sequence
      if ( n < nbIma/2 ) {
        K0 = 1.0 - (1.0-Kmin)*((double)n/(double)(nbIma/2));
      }
      else {
        K0 = Kmin + (1.0-Kmin)*((double)(n-(nbIma/2))/(double)(nbIma/2));
      }
      K1 = 1.0 - K0;
    }
    else
      error();

    printf("POC-%d\tK0=%g", n, K0);
    if ( ficIn[1] ) printf("\tK1=%g\n", K1);
    else            printf("\n");

    setWpScale(&(sc[n]),
      (wpMode==0 || wpMode==2) ? (K0) : (-1.0) ,
      (wpMode==1 || wpMode==2) ? (K1) : (-1.0) );

    byte  *ima0 = ima[0], 
          *ima1 = ima[1], 
          *ima2 = ima[2]; 

    if ( !param->isOption(argc, argv, "fix") || n==0 )
    for ( int i=0 ; i<2 ; i++ )
      if ( ficIn[i] )
        fread(ima[i], sizeof(byte), (nbpLuma+2*nbpChroma), ficIn[i]);
      else
        black(ima[i], sizeX, sizeY);

    // Luma
    for ( int k=0 ; k<nbpLuma ; k++ ) {
      dvalue    = K0*(double)ima0[k] + K1*(double)ima1[k];
      ima2[k]   = (byte)CLAMP(dvalue,0.0,255.0);
    }

    ima0 += nbpLuma;
    ima1 += nbpLuma;
    ima2 += nbpLuma;

    // Chroma
    for ( int uv=0 ; uv<2 ; uv++ ) {
      for ( int k=0 ; k<nbpChroma ; k++ ) {
        dvalue  = 128.0 + K0*((double)ima0[k] - 128.0) + K1*((double)ima1[k] - 128.0);
        ima2[k] = (byte)CLAMP(dvalue,0.0,255.0);
      }
      ima0 += nbpChroma;
      ima1 += nbpChroma;
      ima2 += nbpChroma;
    }

    fwrite(ima[2], sizeof(byte), (nbpLuma+2*nbpChroma), ficOut);
  }


  // Computation of the absolute scaling WP values, depending on the references :
  //-----------------------------------------------------------------------------
  for ( int n=0 ; n<nbIma ; n++ ) {

    if ( ficWpScale && n>0 && n<(nbIma-1) ) {

      if ( strcmp(scenario,"FADE_OUT_FADE_IN")==0 ) {

        //int logWD=7;
        int logWD = computeBestLogWD(sc, n);

        fwriteWpScale_logDW(ficWpScale, logWD, n);
        for ( int eRef=0 ; eRef<2 ; eRef++ ) {
          int iref=0;
          while ( pocRefPicList[n][eRef][iref]>=0 && iref<MAX_NUM_REF ) {
            int ref0 = pocRefPicList[n][eRef][iref];         
            double  K0 = sc[n].K0 / sc[ref0].K0;
            fwriteWpScale1(ficWpScale, logWD, K0, eRef, iref, useOffsetUV, noWeightUV);
            iref++;
          }
        }

      }
      else if ( strcmp(scenario,"CROSS_FADE")==0 ) {

        int logWD=7;
        //int logWD = computeBestLogWD(sc, n);

        fwriteWpScale_logDW(ficWpScale, logWD, n);

        fwriteWpScale1(ficWpScale, logWD, sc[n].K0, 0, 0, useOffsetUV, noWeightUV);
        fwriteWpScale1(ficWpScale, logWD, sc[n].K1, 1, 0, useOffsetUV, noWeightUV);

      }

    }

  }

  if ( ficWpScale ) fclose(ficWpScale);
  if (ficIn[0]) fclose(ficIn[0]);
  if (ficIn[1]) fclose(ficIn[1]);
  fclose(ficOut);
  for ( int i=0 ; i<3 ; i++ )
    free(ima[i]);

}

